"""Tools for interacting with databases in the context of alfred experiments. 

Authors: Christian Treffenstädt (treffenstaedt@psych.uni-goettingen.de) & Johannes Brachem (brachem@psych.uni-goettingen.de)

Alfred3 is a Python library for the generation of dynamic and interactive surveys and experiments. See `alfred3 on GitHub <https://github.com/ctreffe/alfred>`_ for more information on alfred3.
"""


import copy
import pytest
import mongomock
import alfred3 as al

from alfred3_dbtools.mongotools import ExpMongoDBConnector


@pytest.fixture
def exp(tmp_path):
    config = al.config.init_configuration(expdir=tmp_path)
    client = mongomock.MongoClient()
    exp = al.Experiment(config=config)

    exp_agent = al.saving_agent.MongoSavingAgent(
        client=client, database="alfred", collection="exp_col", experiment=exp, name="a1"
    )
    unlinked_agent = al.saving_agent.MongoSavingAgent(
        client=client, database="alfred", collection="unlinked_col", experiment=exp, name="a2"
    )
    codebook_agent = al.saving_agent.MongoSavingAgent(
        client=client, database="alfred", collection="misc_col", experiment=exp, name="a3"
    )

    exp.sac_main.append(exp_agent)
    exp.sac_unlinked.append(unlinked_agent)
    exp.sac_codebook.append(codebook_agent)

    return exp


class TestExpDBConnector:
    def test_init(self, exp):
        db = ExpMongoDBConnector(experiment=exp)
        assert db

    def test_access(self, exp):
        db = ExpMongoDBConnector(experiment=exp)
        assert db.exp_col
        assert db.db
        assert db.unlinked_col
        assert db.misc_col

    def test_multi_agent(self, exp):
        client = mongomock.MongoClient()
        a1b = al.saving_agent.MongoSavingAgent(
            client=client,
            database="alfred",
            collection="exp_col2",
            experiment=exp,
            name="a1b",
            activation_level=1,
        )
        exp.sac_main.append(a1b)
        exp.sac_main.agents["a1"].activation_level = 2

        db = ExpMongoDBConnector(experiment=exp)

        assert db.exp_col.name == "exp_col2"

        a1c = al.saving_agent.MongoSavingAgent(
            client=client,
            database="alfred",
            collection="exp_col3",
            experiment=exp,
            name="a1c",
            activation_level=0.5,
        )

        exp.sac_main.append(a1c)

        db2 = ExpMongoDBConnector(experiment=exp)

        assert db2.exp_col.name == "exp_col3"

    def test_no_agent(self, exp):
        exp.sac_main._agents = {}
        exp.sac_unlinked._agents = {}
        exp.sac_codebook._agents = {}
        db = ExpMongoDBConnector(exp)

        with pytest.raises(ValueError):
            db.exp_col

        with pytest.raises(ValueError):
            db.db

        with pytest.raises(ValueError):
            db.unlinked_col

        with pytest.raises(ValueError):
            db.misc_col



import unittest
import pytest
import os
import copy

import pymongo
import alfred3

from alfred3_dbtools import mongotools


# class Testthepy:

#     @pytest.fixture
#     def mongodb_connector_db(self):
#         _host = os.environ.get("MONGODB_TESTHOST")
#         _port = int(os.environ.get("MONGODB_TESTPORT"))
#         _user1 = os.environ.get("MONGODB_TESTUSER")
#         _pw1 = os.environ.get("MONGODB_TESTPW")
#         _db = os.environ.get("MONGODB_TESTDB")
#         _col = os.environ.get("MONGODB_TESTCOL")
#         _user2 = os.environ.get("MONGODB_TESTUSER2")
#         _pw2 = os.environ.get("MONGODB_TESTPW2")
#         _auth_source = os.environ.get("MONGODB_TEST2AUTH")

#         connector = mongotools.MongoDBConnector(
#             host=_host, port=None, username=_user1, password=_pw1, database=_db
#         )

#         return connector

#     @pytest.fixture
#     def mongodb_connector_col(self):
#         _host = os.environ.get("MONGODB_TESTHOST")
#         _port = int(os.environ.get("MONGODB_TESTPORT"))
#         _user1 = os.environ.get("MONGODB_TESTUSER")
#         _pw1 = os.environ.get("MONGODB_TESTPW")
#         _db = os.environ.get("MONGODB_TESTDB")
#         _col = os.environ.get("MONGODB_TESTCOL")
#         _user2 = os.environ.get("MONGODB_TESTUSER2")
#         _pw2 = os.environ.get("MONGODB_TESTPW2")
#         _auth_source = os.environ.get("MONGODB_TEST2AUTH")

#         connector = mongotools.MongoDBConnector(
#             host=_host, port=None, username=_user1, password=_pw1, database=_db, collection=_col
#         )

#         return connector

#     def test_connect(self, mongodb_connector):
#         con = mongodb_connector
#         con.connect()

#         assert con.connected

#     def test_default(self, mongodb_connector):
#         assert mongodb_connector.connected == False


class TestMongoDBConnector(unittest.TestCase):
    def setUp(self):
        self._host = os.environ.get("MONGODB_TESTHOST")
        self._port = int(os.environ.get("MONGODB_TESTPORT"))
        self._user1 = os.environ.get("MONGODB_TESTUSER")
        self._pw1 = os.environ.get("MONGODB_TESTPW")
        self._db = os.environ.get("MONGODB_TESTDB")
        self._col = os.environ.get("MONGODB_TESTCOL")
        self._user2 = os.environ.get("MONGODB_TESTUSER2")
        self._pw2 = os.environ.get("MONGODB_TESTPW2")
        self._auth_source = os.environ.get("MONGODB_TEST2AUTH")

    def test_minimal_connect(self):
        con = mongotools.MongoDBConnector(
            host=self._host,
            port=None,
            username=self._user1,
            password=self._pw1,
            database=self._db,
        )
        con.connect()
        self.assertTrue(con.connected)
        self.assertIsInstance(con.db, pymongo.database.Database)

    def test_auth_source_connect(self):
        con = mongotools.MongoDBConnector(
            host=self._host,
            port=None,
            username=self._user2,
            password=self._pw2,
            database=self._db,
            auth_source=self._auth_source,
        )
        con.connect()
        self.assertTrue(con.connected)
        self.assertIsInstance(con.db, pymongo.database.Database)

    def test_collection_connect(self):
        con = mongotools.MongoDBConnector(
            host=self._host,
            port=None,
            username=self._user2,
            password=self._pw2,
            database=self._db,
            collection=self._col,
            auth_source=self._auth_source,
        )
        con.connect()
        self.assertTrue(con.connected)
        self.assertIsInstance(con.db, pymongo.collection.Collection)

    def test_autoconnect(self):
        con = mongotools.MongoDBConnector(
            host=self._host,
            port=None,
            username=self._user2,
            password=self._pw2,
            database=self._db,
            collection=self._col,
            auth_source=self._auth_source,
        )
        self.assertFalse(con.connected)
        col = con.db
        self.assertTrue(con.connected)
        print(col)
        self.assertIsInstance(col, pymongo.collection.Collection)

    def test_disconnect(self):
        con = mongotools.MongoDBConnector(
            host=self._host,
            port=None,
            username=self._user2,
            password=self._pw2,
            database=self._db,
            collection=self._col,
            auth_source=self._auth_source,
        )
        con.connect()
        self.assertIsInstance(con.db, pymongo.collection.Collection)
        con.disconnect()
        self.assertFalse(con.connected)

    def test_insert_find_remove(self):
        con = mongotools.MongoDBConnector(
            host=self._host,
            port=None,
            username=self._user2,
            password=self._pw2,
            database=self._db,
            collection=self._col,
            auth_source=self._auth_source,
        )
        col = con.db
        test_document = {"test": "test"}

        col.insert_one(test_document)
        query = col.find_one(test_document)
        self.assertIsNotNone(query)

        col.delete_one(test_document)
        delquery = col.find_one(test_document)
        self.assertIsNone(delquery)


if __name__ == "__main__":
    unittest.main()



"""Tools for interacting with MongoDB databases.
"""

import copy

from pymongo import MongoClient
import alfred3


class MongoDBConnector:
    """Connect to a MongoDB database or collection.

    Effectively, this class provides an interface to an instace of ``pymongo.MongoClient``. For further information on how to interact with a `MongoClient`, please refer to the `pymongo documentation <https://pymongo.readthedocs.io/en/stable/tutorial.html>`_.
    
    :param str host: Hostname or IP address.
    :param int port: Port number on which to connect.
    :param str username: Username for authentication.
    :param str password: Password for authentication.
    :param str database: Database to which to the client will connect.
    :param str collection: Collection inside the specified database to which the client will connect. If left empty, the client will connect directly to the database.
    :param str auth_source: Database to authenticate on. Defaults to "admin", the MongoDB default.
    :param bool ssl: If ``True``, create a connection to the server using Transport Layer Security (TLS/SSL).
    :param str ca_file: Filepath to a file containing a single or a bundle of “certification authority” certificates, which are used to validate certificates passed from the other end of the connection. Implies ``ssl=True``. Defaults to ``None``.
    """

    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        password: str,
        database: str,
        collection: str = None,
        auth_source: str = "admin",
        ssl: bool = False,
        ca_file: str = None,
    ):
        """Constructor method."""

        self._host = host
        self._port = port
        self._db_name = database
        self._collection = collection
        self._username = username
        self._password = password
        self._auth_source = auth_source
        self._ssl = True if ca_file else ssl
        self._ca_file = ca_file

        self._client = None
        self._db = None

        self.connected = False
        """`True`, if a connection was established."""

        self.connect()

    def connect(self):
        """Establish the connection the MongoDB.

        If a collection was specified upon initialisation, the `MongoDBConnector` will connect to this collection. Else, it will connect to the specified database.
        """

        self._client = MongoClient(
            host=self._host,
            port=self._port,
            username=self._username,
            password=self._password,
            authSource=self._auth_source,
            tls=self._ssl,
            tlsCAFile=self._ca_file,
        )

        db = self._client[self._db_name]

        if self._collection:
            db = db[self._collection]

        self._db = db
        self.connected = True

    def disconnect(self):
        """Close the connection to the database."""

        self._client.close()
        self.connected = False

    @property
    def db(self):
        """Return the database or, if specified, collection given upon initialisation.
        
        If the instance of ``MongoDBConnector`` is not currently connected, a call to this property will trigger a call to ``MongoDBConnector.connect()`` before returning the database or collection.
        """

        if self.connected:
            if self._db is None:
                raise TypeError("No collection found.")
            else:
                return self._db

        elif not self.connected:
            self.connect()
            if self._db is None:
                raise TypeError("No collection found.")
            else:
                return self._db


class ExpMongoDBConnector:
    """Connect to an alfred experiment's MongoDB collection.
    
    Connects to the same database and collection that the given alfred 
    experiment uses to save its data. If multiple `MongoSavingAgents` 
    are attached to the experiment, :class:`ExpMongoDBConnector` will connect
    to the `MongoSavingAgent` with the lowest activation level. 
    Basically, it provides access to a copy of the
    :class:`pymongo.collection.Collection` that the experiment's saving agent
    uses. 

    For further information on how to interact with a
    `MongoClient`, please refer to the 
    `pymongo documentation <https://pymongo.readthedocs.io/en/stable/tutorial.html>`_.

    Args:
        experiment: An alfred3 experiment instance.
    """

    def __init__(self, experiment: alfred3.Experiment):
        """Constructor method."""

        self._exp = experiment

        if not isinstance(self._exp, alfred3.Experiment):
            raise ValueError("The input must be an instance of alfred3.Experiment.")

        self.exp_data_agents = []
        self.unlinked_data_agents = []
        self.codebook_agents = []

        self._gather_agents(self.exp_data_agents, self._exp.sac_main)
        self._gather_agents(self.unlinked_data_agents, self._exp.sac_unlinked)
        self._gather_agents(self.codebook_agents, self._exp.sac_codebook)

        self._exp_collection = self._connect(self.exp_data_agents)
        self._unlinked_collection = self._connect(self.unlinked_data_agents)
        self._misc_collection = self._connect(self.codebook_agents)

    def _gather_agents(self, agent_list, sac):
        """Collect all MongoSavingAgents from the provided alfred 
        experiment, sorted by activation level (lowest first).
        
        Args:
            agent_list: The list in which saving agents should be 
                collected.
            sac: The SavingAgentController, from which saving agents
                should be collected.
        """

        for agent in sac.agents.values():
            if isinstance(agent, alfred3.saving_agent.MongoSavingAgent):
                agent_list.append(copy.copy(agent))
        agent_list.sort(key=lambda x: x.activation_level)

    def _connect(self, agent_list):
        """Establishes a connection to the MongoDB collection with the 
        lowest activation level in *agent_list*."""

        if (
            len(agent_list) > 1
            and agent_list[0].activation_level == agent_list[1].activation_level
        ):
            raise ValueError(
                "There are two or more MongoSavingAgents with the highest activation level."
            )

        try:
            return agent_list[0]._col
        except IndexError:
            pass

    @property
    def unlinked_col(self):
        """Returns the unlinked mongoDB collection belonging to the
        experiment.
        """
        if not self._unlinked_collection:
            raise ValueError("No unlinked collection found.")
        return self._unlinked_collection

    @property
    def misc_col(self):
        """Returns the miscellaneous mongoDB collection belonging to the
        experiment.
        """
        if not self._misc_collection:
            raise ValueError("No miscellaneous collection found.")
        return self._misc_collection

    @property
    def db(self):
        """Returns the mongoDB collection from the experiment's 
        ``MongoSavingAgent`` with lowest activation level.
        """
        DeprecationWarning("This property is deprecated. Please use 'exp' instead.")
        if not self._exp_collection:
            raise ValueError("No experiment data collection found.")
        return self._exp_collection

    @property
    def exp_col(self):
        """Returns the mongoDB collection from the experiment's 
        ``MongoSavingAgent`` with lowest activation level.
        """
        if not self._exp_collection:
            raise ValueError("No experiment data collection found.")
        return self._exp_collection



# -*- coding: utf-8 -*-

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

# Parse version from _version.py in package directory
# See https://packaging.python.org/guides/single-sourcing-package-version/#single-sourcing-the-version
version = {}
with open("alfred3_dbtools/_version.py") as f:
    exec(f.read(), version)

setuptools.setup(
    name="alfred3_dbtools",
    version=version["__version__"],
    author="Christian Treffenstädt, Johannes Brachem",
    author_email="treffenstaedt@psych.uni-goettingen.de",
    description="A package that provides tools for interacting with databases when working with alfred3 experiments.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ctreffe/alfred3-dbtools",
    packages=setuptools.find_packages(),
    install_requires=["alfred3>=1.1.4", "pymongo>=3.10"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)
